/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include	"../quakedef.h"
#include	"../gl_gentex.h"
#include	"../menu.h"

#define	MAX_MOD_KNOWN	256

extern	gentex_t gentex[NUM_GENTEXES];

model_t	*loadmodel;
model_t *Mod_LoadModel (model_t *mod, qboolean crash);
model_t	mod_known[MAX_MOD_KNOWN];

int		mod_numknown;
char	loadname[32];
byte	mod_novis[MAX_MAP_LEAFS/8];

void Mod_LoadBrushModel (model_t *mod, void *buffer);

/*
========
Mod_Init
========
*/
void Mod_Init (void)
{
	MR_memset (mod_novis, 0xff, sizeof(mod_novis));
}

/*
=============
Mod_Extradata
=============
*/
void *Mod_Extradata (model_t *mod)
{
	void	*r;

	r = Cache_Check (&mod->cache);
	if (r)
		return r;

	Mod_LoadModel (mod, true);
	
	if (!mod->cache.data)
		Sys_Error ("Mod_Extradata: caching failed");
	return mod->cache.data;
}

/*
===============
Mod_PointInLeaf
===============
*/
mleaf_t *Mod_PointInLeaf (vec3_t p, model_t *model)
{
	mnode_t		*node;
	float		d;
	mplane_t	*plane;
	
	if (!model || !model->nodes)
		Sys_Error ("Mod_PointInLeaf: bad model");

	node = model->nodes;
	while (1)
	{
		if (node->contents < 0)
			return (mleaf_t *)node;
		plane = node->plane;
		d = PlaneDiff(p, plane);
		if (d > 0)
			node = node->children[0];
		else
			node = node->children[1];
	}
	
	return NULL;	// never reached
}

/*
=================
Mod_DecompressVis
=================
*/
byte *Mod_DecompressVis (byte *in, model_t *model)
{
	static byte	decompressed[MAX_MAP_LEAFS/8];
	int		c;
	byte	*out;
	int		row;

	row = (model->numleafs+7)>>3;	
	out = decompressed;

	if (!in)
	{	// no vis info, so make all visible
		while (row)
		{
			*out++ = 0xff;
			row--;
		}
		return decompressed;		
	}

	do
	{
		if (*in)
		{
			*out++ = *in++;
			continue;
		}
	
		c = in[1];
		in += 2;
		while (c)
		{
			*out++ = 0;
			c--;
		}
	} while (out - decompressed < row);
	
	return decompressed;
}

byte *Mod_LeafPVS (mleaf_t *leaf, model_t *model)
{
	if (leaf == model->leafs)
		return mod_novis;
	return Mod_DecompressVis (leaf->compressed_vis, model);
}

/*
============
Mod_ClearAll
============
*/
void Mod_ClearAll (void)
{
	model_t	*mod;
	int		i;
	
	for (i=0 , mod=mod_known ; i<mod_numknown ; i++, mod++)
		if (mod->type != mod_car)
			mod->needload = true;
}

/*
============
Mod_FindName
============
*/
model_t *Mod_FindName (char *name)
{
	model_t	*mod;
	int		i;
	
	if (!name[0])
		Sys_Error ("Mod_ForName: NULL name");

	for (i=0 , mod=mod_known ; i<mod_numknown ; i++, mod++)
		if (!MR_strcmp (mod->name, name) )
			break;
			
	if (i == mod_numknown)
	{
		if (mod_numknown == MAX_MOD_KNOWN)
			Sys_Error ("mod_numknown == MAX_MOD_KNOWN");
		MR_strcpy (mod->name, name);
		mod->needload = true;
		mod_numknown++;
	}

	return mod;
}

/*
=============
Mod_LoadModel
=============
*/
void Mod_LoadCARModel (model_t *mod, void *buffer);

#define CARHEADER 666

model_t *Mod_LoadModel (model_t *mod, qboolean crash)
{
	unsigned	*buf;
	byte		stackbuf[1024];		// avoid dirtying the cache heap

	if (!mod->needload)
		return mod;
	
//
// load the file
//
	buf = (unsigned *)COM_LoadStackFile (mod->name, stackbuf, sizeof(stackbuf));
	if (!buf)
	{
		if (crash)
			Sys_Error ("Mod_NumForName: %s not found", mod->name);
		return NULL;
	}
	
//
// allocate a new model
//
	COM_FileBase (mod->name, loadname);
	
	loadmodel = mod;

//
// fill it in
//

// call the apropriate loader
	mod->needload = false;
	
	switch (LittleLong(*(unsigned *)buf))
	{
	case CARHEADER: 
		Mod_LoadCARModel (mod, buf); 
		break;

	default:
		Mod_LoadBrushModel (mod, buf);
		break;
	}

	return mod;
}

/*
===========
Mod_ForName
===========
*/
model_t *Mod_ForName (char *name, qboolean crash)
{
	model_t	*mod;
	
	mod = Mod_FindName (name);
	
	return Mod_LoadModel (mod, crash);
}

/*
===============================================================================

					BRUSHMODEL LOADING

===============================================================================
*/

byte *mod_base;

byte *LoadMRTData (char* filename, qboolean complain);
byte *GenerateTexture (gentex_t *gentex);
void GenerateTextureStart (char *scene);
void GenerateTextureStop  (void);

/*
================
Mod_LoadTextures
================
*/
void Mod_LoadTextures (lump_t *l)
{
	int				i, j, k, *dofs;
	miptex_t		*mt;
	texture_t		*tx;
	dmiptexlump_t	*m;
	byte			*data = NULL;
	char			tex_name[32];
	qboolean		loaded;

	extern	int		image_width;
	extern	int		image_height;
	extern	int		current_theme;
	extern	list_t	themes[64];

	if (!l->filelen)
	{
		loadmodel->textures = NULL;
		return;
	}

	m						= (dmiptexlump_t *)(mod_base + l->fileofs);
	m->nummiptex			= LittleLong (m->nummiptex);
	loadmodel->numtextures	= m->nummiptex;
	loadmodel->textures		= Hunk_AllocName (m->nummiptex * sizeof(*loadmodel->textures) , loadname);
	dofs					= m->dataofs;

	GenerateTextureStart (themes[current_theme].name);

	Con_Printf ("Loading textures");

	for (i=0 ; i<m->nummiptex ; i++)
	{
		dofs[i] = LittleLong (dofs[i]);

		if (dofs[i])
		{
			mt			= (miptex_t *)((byte *)m + dofs[i]);
			mt->width	= LittleLong (mt->width);
			mt->height	= LittleLong (mt->height);

			if ((mt->width & 15) || (mt->height & 15))
				Host_Error ("Texture %s is not 16 aligned", mt->name);

			tx = Hunk_AllocName (sizeof(texture_t), loadname);
			loadmodel->textures[i] = tx;

			for (j = 0;mt->name[j] && j < 15;j++)
			{
				if (mt->name[j] >= 'A' && mt->name[j] <= 'Z')
					tx->name[j] = mt->name[j] + ('a' - 'A');
				else
					tx->name[j] = mt->name[j];
			}

			tx->width	= mt->width;
			tx->height	= mt->height;

			if ((!MR_strncmp (tx->name, "sky", 3))		||
				(!MR_strncmp (tx->name, "sign", 4))		||
				(!MR_strncmp (tx->name, "water", 5)))
				continue;

			loaded = false;

			for (k=0; k<NUM_GENTEXES; k++)
			{
				if (!MR_strcmp (gentex[k].name, tx->name))
				{
					data = GenerateTexture (&gentex[k]);
					if (data)
					{
						loaded = true;
						image_width = image_height = 256;
					}
				}
			}

			if(!loaded)
			{
				_snprintf (tex_name, sizeof(tex_name), "textures/%s/%s", themes[current_theme].name, tx->name);
				data = LoadMRTData (tex_name, false);
				if (data)
					loaded = true;
			}

			if(!loaded)
			{
				_snprintf (tex_name, sizeof(tex_name), "textures/%s", tx->name);
				data = LoadMRTData (tex_name, true);
				if (data)
					loaded = true;
			}

			if(!loaded)
				continue;

			tx->gl_texturenum = GL_LoadTexture (tx->name, image_width, image_height, data, true, false);
			
			free (data);

			Con_Printf (".");
		}
	}

	GenerateTextureStop ();

	Con_Printf ("\nDone\n");
}

/*
================
Mod_LoadLighting
================
*/
void Mod_LoadLighting (lump_t *l, int num)
{
	int				i;
	byte			*in, *out;
	byte			d;
	float			red = 1, green = 1, blue = 1;

	extern cvar_t	_timeofday;

	if (!l->filelen)
		return;

	if (num != _timeofday.ivalue)
		return;

	loadmodel->lightdata = Hunk_AllocName (l->filelen*3, loadmodel->name);

	in	= loadmodel->lightdata + l->filelen*2;
	out = loadmodel->lightdata;

	MR_memcpy (in, mod_base + l->fileofs, l->filelen);

	switch (_timeofday.ivalue)
	{
	case 1:
		blue	= 0.75;
		break;

	case 3:
		green	= 0.75;
		blue	= 0.75;
		break;
	}

	for (i = 0;i < l->filelen;i++)
	{
		d		= *in;
		out[0]	= d * red;
		out[1]	= d * green;
		out[2]	= d * blue;

		in		+= 1;
		out		+= 3;
	}

	Con_Printf ("Loaded lighting %i\n", num);
}

/*
==================
Mod_LoadVisibility
==================
*/
void Mod_LoadVisibility (lump_t *l)
{
	if (!l->filelen)
	{
		loadmodel->visdata = NULL;
		return;
	}
	loadmodel->visdata = Hunk_AllocName ( l->filelen, loadname);	
	MR_memcpy (loadmodel->visdata, mod_base + l->fileofs, l->filelen);
}

/*
================
Mod_LoadEntities
================
*/
void Mod_LoadEntities (lump_t *l)
{
	if (!l->filelen)
	{
		loadmodel->entities = NULL;
		return;
	}
	loadmodel->entities = Hunk_AllocName ( l->filelen, loadname);	
	MR_memcpy (loadmodel->entities, mod_base + l->fileofs, l->filelen);
}

/*
================
Mod_LoadVertexes
================
*/
void Mod_LoadVertexes (lump_t *l)
{
	dvertex_t	*in;
	mvertex_t	*out;
	int			i, count;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->vertexes		= out;
	loadmodel->numvertexes	= count;

	for (i=0; i<count; i++, in++, out++)
	{
		out->position[0] = LittleFloat (in->point[0]);
		out->position[1] = LittleFloat (in->point[1]);
		out->position[2] = LittleFloat (in->point[2]);
	}
}

/*
=================
Mod_LoadSubmodels
=================
*/
void Mod_LoadSubmodels (lump_t *l)
{
	dmodel_t	*in;
	dmodel_t	*out;
	int			i, j, count;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->submodels = out;
	loadmodel->numsubmodels = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		for (j=0 ; j<3 ; j++)
		{	// spread the mins / maxs by a pixel
			out->mins[j] = LittleFloat (in->mins[j]) - 1;
			out->maxs[j] = LittleFloat (in->maxs[j]) + 1;
			out->origin[j] = LittleFloat (in->origin[j]);
		}
		for (j=0 ; j<MAX_MAP_HULLS ; j++)
			out->headnode[j] = LittleLong (in->headnode[j]);
		out->visleafs = LittleLong (in->visleafs);
		out->firstface = LittleLong (in->firstface);
		out->numfaces = LittleLong (in->numfaces);
	}
}

/*
=============
Mod_LoadEdges
=============
*/
void Mod_LoadEdges (lump_t *l)
{
	dedge_t *in;
	medge_t *out;
	int 	i, count;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( (count + 1) * sizeof(*out), loadname);	

	loadmodel->edges = out;
	loadmodel->numedges = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		out->v[0] = (unsigned short)LittleShort(in->v[0]);
		out->v[1] = (unsigned short)LittleShort(in->v[1]);
	}
}

/*
===============
Mod_LoadTexinfo
===============
*/
void Mod_LoadTexinfo (lump_t *l)
{
	texinfo_t *in;
	mtexinfo_t *out;
	int 	i, j, count;
	int		miptex;
	float	len1, len2;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->texinfo = out;
	loadmodel->numtexinfo = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		for (j=0 ; j<8 ; j++)
			out->vecs[0][j] = LittleFloat (in->vecs[0][j]);
		len1 = Length (out->vecs[0]);
		len2 = Length (out->vecs[1]);
		len1 = (len1 + len2)/2;
		if (len1 < 0.32)
			out->mipadjust = 4;
		else if (len1 < 0.49)
			out->mipadjust = 3;
		else if (len1 < 0.99)
			out->mipadjust = 2;
		else
			out->mipadjust = 1;

		miptex = LittleLong (in->miptex);
		out->flags = LittleLong (in->flags);
	
		if (!loadmodel->textures)
		{
			Con_Printf ("WARNING: No texture found");
			out->texture	= 0;
			out->flags		= 0;
		}
		else
		{
			if (miptex >= loadmodel->numtextures)
				Sys_Error ("miptex >= loadmodel->numtextures");
			out->texture = loadmodel->textures[miptex];
			if (!out->texture)
			{
				Con_Printf ("WARNING: No texture found");
				out->texture	= 0;
				out->flags		= 0;
			}
		}
	}
}

/*
==================
CalcSurfaceExtents
==================
*/
void CalcSurfaceExtents (msurface_t *s)
{
	mtexinfo_t	*tex;
	mvertex_t	*v;
	float		mins[2], maxs[2], val;
	int			i,j, e, bmins[2], bmaxs[2];

	mins[0] = mins[1] =  99999;
	maxs[0] = maxs[1] = -99999;

	tex = s->texinfo;
	
	for (i=0 ; i<s->numedges ; i++)
	{
		e = loadmodel->surfedges[s->firstedge+i];
		if (e >= 0)
			v = &loadmodel->vertexes[loadmodel->edges[e].v[0]];
		else
			v = &loadmodel->vertexes[loadmodel->edges[-e].v[1]];
		
		for (j=0 ; j<2 ; j++)
		{
			val = v->position[0] * tex->vecs[j][0] + v->position[1] * tex->vecs[j][1] +	v->position[2] * tex->vecs[j][2] + tex->vecs[j][3];
		
			if (val < mins[j])
				mins[j] = val;
			if (val > maxs[j])
				maxs[j] = val;
		}
	}

	for (i=0 ; i<2 ; i++)
	{	
		bmins[i] = MR_floor(mins[i] * 0.0625);	// Speed... were ( / 16)
		bmaxs[i] = MR_ceil (maxs[i] * 0.0625);	// Speed... were ( / 16)

		s->texturemins[i] = bmins[i] * 16;
		s->extents[i] = (bmaxs[i] - bmins[i]) * 16;
		if (!(tex->flags & TEX_SPECIAL) && s->extents[i] > 2032)
			Host_Error ("Bad surface extents %i", s->extents[i]);
	}
}

/*
=============
Mod_LoadFaces
=============
*/
void Mod_LoadFaces (lump_t *l)
{
	dface_t		*in;
	msurface_t 	*out;
	int			i, count, surfnum;
	int			planenum, side;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->surfaces = out;
	loadmodel->numsurfaces = count;

	for ( surfnum=0 ; surfnum<count ; surfnum++, in++, out++)
	{
		out->firstedge = LittleLong(in->firstedge);
		out->numedges = LittleShort(in->numedges);		
		out->flags = 0;

		planenum = LittleShort(in->planenum);
		side = LittleShort(in->side);
		if (side)
			out->flags |= SURF_PLANEBACK;			

		out->plane = loadmodel->planes + planenum;

		out->texinfo = loadmodel->texinfo + LittleShort (in->texinfo);

		CalcSurfaceExtents (out);
				
	// lighting info

		i = LittleLong(in->lightofs);
		if (i == -1)
			out->samples = NULL;
		else
			out->samples = loadmodel->lightdata + (i * 3);

	// set the drawing flags flag

		if ((out->texinfo->texture->name[0] == 's') &&
			(out->texinfo->texture->name[1] == 'k') &&
			(out->texinfo->texture->name[2] == 'y') &&
			(out->texinfo->texture->name[3] == 'u') &&
			(out->texinfo->texture->name[4] == 'p'))
		{
			out->flags |= SURF_DRAWSKYUP;
			continue;
		}		
		if ((out->texinfo->texture->name[0] == 's') &&
			(out->texinfo->texture->name[1] == 'k') &&
			(out->texinfo->texture->name[2] == 'y'))
		{
			out->flags |= SURF_DRAWSKY;
			continue;
		}
		if ((out->texinfo->texture->name[0] == 'w') &&
			(out->texinfo->texture->name[1] == 'a') &&
			(out->texinfo->texture->name[2] == 't') &&
			(out->texinfo->texture->name[3] == 'e') &&
			(out->texinfo->texture->name[4] == 'r'))
		{
			out->flags |= SURF_DRAWWATER;
			continue;
		}

		if ((out->texinfo->texture->name[0] == 's') &&
			(out->texinfo->texture->name[1] == 'i') &&
			(out->texinfo->texture->name[2] == 'g') &&
			(out->texinfo->texture->name[3] == 'n'))
		{
			out->flags |= SURF_DRAWSIGN;
			continue;
		}
	}
}

/*
=============
Mod_SetParent
=============
*/
void Mod_SetParent (mnode_t *node, mnode_t *parent)
{
	node->parent = parent;
	if (node->contents < 0)
		return;
	Mod_SetParent (node->children[0], node);
	Mod_SetParent (node->children[1], node);
}

/*
=============
Mod_LoadNodes
=============
*/
void Mod_LoadNodes (lump_t *l)
{
	int			i, j, count, p;
	dnode_t		*in;
	mnode_t 	*out;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->nodes = out;
	loadmodel->numnodes = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		for (j=0 ; j<3 ; j++)
		{
			out->mins[j] = LittleShort (in->mins[j]);
			out->maxs[j] = LittleShort (in->maxs[j]);
		}
	
		p = LittleLong(in->planenum);
		out->plane = loadmodel->planes + p;

		out->firstsurface = LittleShort (in->firstface);
		out->numsurfaces = LittleShort (in->numfaces);
		
		for (j=0 ; j<2 ; j++)
		{
			p = LittleShort (in->children[j]);
			if (p >= 0)
				out->children[j] = loadmodel->nodes + p;
			else
				out->children[j] = (mnode_t *)(loadmodel->leafs + (-1 - p));
		}
	}
	
	Mod_SetParent (loadmodel->nodes, NULL);	// sets nodes and leafs
}

/*
=============
Mod_LoadLeafs
=============
*/
void Mod_LoadLeafs (lump_t *l)
{
	dleaf_t 	*in;
	mleaf_t 	*out;
	int			i, j, count, p;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->leafs = out;
	loadmodel->numleafs = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		for (j=0 ; j<3 ; j++)
		{
			out->mins[j] = LittleShort (in->mins[j]);
			out->maxs[j] = LittleShort (in->maxs[j]);
		}

		p = LittleLong(in->contents);
		out->contents = p;

		out->firstmarksurface = loadmodel->marksurfaces +
			LittleShort(in->firstmarksurface);
		out->nummarksurfaces = LittleShort(in->nummarksurfaces);
		
		p = LittleLong(in->visofs);
		if (p == -1)
			out->compressed_vis = NULL;
		else
			out->compressed_vis = loadmodel->visdata + p;
	}	
}

/*
=================
Mod_LoadClipnodes
=================
*/
void Mod_LoadClipnodes (lump_t *l)
{
	dclipnode_t *in, *out;
	int			i, count;
	hull_t		*hull;

	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->clipnodes = out;
	loadmodel->numclipnodes = count;

	hull				= &loadmodel->hulls[1];
	hull->clipnodes		= out;
	hull->firstclipnode = 0;
	hull->lastclipnode	= count-1;
	hull->planes		= loadmodel->planes;
	hull->clip_mins[0]	= -12;
	hull->clip_mins[1]	= -12;
	hull->clip_mins[2]	= -4;
	hull->clip_maxs[0]	= 12;
	hull->clip_maxs[1]	= 12;
	hull->clip_maxs[2]	= 4;

	hull				= &loadmodel->hulls[2];
	hull->clipnodes		= out;
	hull->firstclipnode = 0;
	hull->lastclipnode	= count-1;
	hull->planes		= loadmodel->planes;
	hull->clip_mins[0]	= -12;
	hull->clip_mins[1]	= -12;
	hull->clip_mins[2]	= -4;
	hull->clip_maxs[0]	= 12;
	hull->clip_maxs[1]	= 12;
	hull->clip_maxs[2]	= 4;

	for (i=0 ; i<count ; i++, out++, in++)
	{
		out->planenum = LittleLong(in->planenum);
		out->children[0] = LittleShort(in->children[0]);
		out->children[1] = LittleShort(in->children[1]);
	}
}

/*
=============
Mod_MakeHull0
=============
*/
void Mod_MakeHull0 (void)
{
	mnode_t		*in, *child;
	dclipnode_t *out;
	int			i, j, count;
	hull_t		*hull;
	
	hull = &loadmodel->hulls[0];	
	
	in = loadmodel->nodes;
	count = loadmodel->numnodes;
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	hull->clipnodes = out;
	hull->firstclipnode = 0;
	hull->lastclipnode = count-1;
	hull->planes = loadmodel->planes;

	for (i=0 ; i<count ; i++, out++, in++)
	{
		out->planenum = in->plane - loadmodel->planes;
		for (j=0 ; j<2 ; j++)
		{
			child = in->children[j];
			if (child->contents < 0)
				out->children[j] = child->contents;
			else
				out->children[j] = child - loadmodel->nodes;
		}
	}
}

/*
====================
Mod_LoadMarksurfaces
====================
*/
void Mod_LoadMarksurfaces (lump_t *l)
{	
	int		i, j, count;
	short		*in;
	msurface_t **out;
	
	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->marksurfaces = out;
	loadmodel->nummarksurfaces = count;

	for ( i=0 ; i<count ; i++)
	{
		j = LittleShort(in[i]);
		if (j >= loadmodel->numsurfaces)
			Sys_Error ("Mod_ParseMarksurfaces: bad surface number");
		out[i] = loadmodel->surfaces + j;
	}
}

/*
=================
Mod_LoadSurfedges
=================
*/
void Mod_LoadSurfedges (lump_t *l)
{	
	int		i, count;
	int		*in, *out;
	
	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*sizeof(*out), loadname);	

	loadmodel->surfedges = out;
	loadmodel->numsurfedges = count;

	for ( i=0 ; i<count ; i++)
		out[i] = LittleLong (in[i]);
}

/*
==============
Mod_LoadPlanes
==============
*/
void Mod_LoadPlanes (lump_t *l)
{
	int			i, j;
	mplane_t	*out;
	dplane_t 	*in;
	int			count;
	int			bits;
	
	in = (void *)(mod_base + l->fileofs);
	if (l->filelen % sizeof(*in))
		Sys_Error ("MOD_LoadBmodel: funny lump size in %s",loadmodel->name);
	count = l->filelen / sizeof(*in);
	out = Hunk_AllocName ( count*2*sizeof(*out), loadname);	
	
	loadmodel->planes = out;
	loadmodel->numplanes = count;

	for ( i=0 ; i<count ; i++, in++, out++)
	{
		bits = 0;
		for (j=0 ; j<3 ; j++)
		{
			out->normal[j] = LittleFloat (in->normal[j]);
			if (out->normal[j] < 0)
				bits |= 1<<j;
		}

		out->dist = LittleFloat (in->dist);
		out->type = LittleLong (in->type);
		BoxOnPlaneSideClassify (out);
	}
}

/*
================
RadiusFromBounds
================
*/
float RadiusFromBounds (vec3_t mins, vec3_t maxs)
{
	int		i;
	vec3_t	corner;

	for (i=0 ; i<3 ; i++)
	{
		corner[i] = MR_fabs(mins[i]) > MR_fabs(maxs[i]) ? MR_fabs(mins[i]) : MR_fabs(maxs[i]);
	}

	return Length (corner);
}

/*
==================
Mod_LoadBrushModel
==================
*/
void Mod_LoadBrushModel (model_t *mod, void *buffer)
{
	int			i, j;
	dheader_t	*header;
	dmodel_t 	*bm;
	
	loadmodel->type = mod_brush;
	
	header = (dheader_t *)buffer;

	i = LittleLong (header->version);

	if (i != MRBSP)
	{
		Con_Printf("Mod_LoadBrushModel: %s has wrong version number (%i should be %i)", mod->name, i, MRBSP);
		mod->numvertexes = -1;
		return;
	}

// swap all the lumps
	mod_base = (byte *)header;

	for (i=0 ; i<sizeof(dheader_t)/4 ; i++)
		((int *)header)[i] = LittleLong ( ((int *)header)[i]);

// load into heap
	
	Mod_LoadVertexes		(&header->lumps[LUMP_VERTEXES]);
	Mod_LoadEdges			(&header->lumps[LUMP_EDGES]);
	Mod_LoadSurfedges		(&header->lumps[LUMP_SURFEDGES]);
	if(cls.state != ca_dedicated)
	{
	    Mod_LoadTextures		(&header->lumps[LUMP_TEXTURES]);
	    Mod_LoadLighting		(&header->lumps[LUMP_LIGHTING0], 0);
	    Mod_LoadLighting		(&header->lumps[LUMP_LIGHTING1], 1);
	    Mod_LoadLighting		(&header->lumps[LUMP_LIGHTING2], 2);
	    Mod_LoadLighting		(&header->lumps[LUMP_LIGHTING3], 3);
	}
	Mod_LoadPlanes			(&header->lumps[LUMP_PLANES]);
	if(cls.state != ca_dedicated)
	{
	    Mod_LoadTexinfo			(&header->lumps[LUMP_TEXINFO]);
	    Mod_LoadFaces			(&header->lumps[LUMP_FACES]);
	    Mod_LoadMarksurfaces	(&header->lumps[LUMP_MARKSURFACES]);
	    Mod_LoadVisibility		(&header->lumps[LUMP_VISIBILITY]);
	}
	Mod_LoadLeafs			(&header->lumps[LUMP_LEAFS]);
	Mod_LoadNodes			(&header->lumps[LUMP_NODES]);
	Mod_LoadClipnodes		(&header->lumps[LUMP_CLIPNODES]);
	Mod_LoadEntities		(&header->lumps[LUMP_ENTITIES]);
	Mod_LoadSubmodels		(&header->lumps[LUMP_MODELS]);

	Mod_MakeHull0 ();
	
	mod->numframes = 2;		// regular and alternate animation
	
//
// set up the submodels (FIXME: this is confusing)
//
	for (i=0 ; i<mod->numsubmodels ; i++)
	{
		bm = &mod->submodels[i];

		mod->hulls[0].firstclipnode = bm->headnode[0];
		for (j=1 ; j<MAX_MAP_HULLS ; j++)
		{
			mod->hulls[j].firstclipnode	= bm->headnode[j];
			mod->hulls[j].lastclipnode	= mod->numclipnodes-1;
		}
		
		mod->firstmodelsurface	= bm->firstface;
		mod->nummodelsurfaces	= bm->numfaces;
		
		VectorCopy (bm->maxs, mod->maxs);
		VectorCopy (bm->mins, mod->mins);

		mod->radius = RadiusFromBounds (mod->mins, mod->maxs);

		mod->numleafs = bm->visleafs;

		if (i < mod->numsubmodels-1)
		{	// duplicate the basic information
			char	name[10];

			_snprintf (name, sizeof(name), "*%i", i+1);
			loadmodel = Mod_FindName (name);
			*loadmodel = *mod;
			MR_strcpy (loadmodel->name, name);
			mod = loadmodel;
		}
	}
}
